#include <wos.hpp>

/////////////////////////////////////////////
//// Helper functions, defined below.
/////////////////////////////////////////////
// Generates a point evenly placed on a slice plane along the xy.
std::vector<wos::Vector3> generate_slice_plane_points(wos::Vector3 min, wos::Vector3 max, float z, int width, int height);

// Computes a bound of the transformed absorption coefficient of the PDE.
// Here this is done approximately using random sampling, which is not ideal.
// It does work reasonably well in practice and all the results in the paper use a similar approach.
template<typename T, int DIM>
float compute_bound(const wos::PDE<T, DIM> pde, const wos::Vector<DIM>& bmin, const wos::Vector<DIM>& bmax);

// Evaluates the distance to the "bunny" domain.
float domain_bunny(wos::Vector3 p);

// Colormap for visualization.
wos::Vector3 viridis(float t);
/////////////////////////////////////////////


int main(void) {
    /////////////////////////////////////////////
    // Set up our problem description.
    /////////////////////////////////////////////
    wos::PDE<float, 3> pde;
    {
        float absorptionMin = 10;
        float absorptionMax = 100;
        float diffusionFreq = 0.5;
        float dirichletFreq = 1.5;

        pde.mediumType = wos::MediumType::Absorbing;
        pde.absorptionFunc = [absorptionMin, absorptionMax](const wos::Vector3& x) -> float {
            return absorptionMin + (absorptionMax - absorptionMin) * (1.0f + 0.5f*std::sin(2.0f*wos::M_PI*x(0))*std::cos(0.5f*wos::M_PI*x(1)));
        };
        pde.diffusionFunc = [diffusionFreq](const wos::Vector3& x, 
            bool computeGradient, bool computeLaplacian, wos::Vector3& gradient, float& laplacian) -> float {
            float a = 4.0f*wos::M_PI*diffusionFreq;
            float b = 3.0f*wos::M_PI*diffusionFreq;
            float sinAx = std::sin(a*x(0));
            float cosAx = std::cos(a*x(0));
            float sinBy = std::sin(b*x(1));
            float cosBy = std::cos(b*x(1));
            float alpha = std::exp(-x(1)*x(1) + cosAx*sinBy);

            if (computeGradient || computeLaplacian) {
                gradient(0) = alpha*(-sinAx*sinBy*a);
                gradient(1) = alpha*(-2.0f*x(1) + cosAx*cosBy*b);
            }

            if (computeLaplacian) {
                float d2Alphadx2 = gradient(0)*(-sinAx*sinBy*a) + alpha*(-cosAx*sinBy*a*a);
                float d2Alphady2 = gradient(1)*(-2.0f*x(1) + cosAx*cosBy*b) + alpha*(-2.0f - cosAx*sinBy*b*b);
                laplacian = d2Alphadx2 + d2Alphady2;
            }

            return alpha;
        };
        pde.dirichletFunc = [dirichletFreq](const wos::Vector3& x) -> float {
            float k = wos::M_PI*dirichletFreq;
            float sinAx = std::sin(k*x(0));
            float cosAx = std::cos(k*x(0));
            float sinBy = std::sin(2.0f*k*x(1));
            float cosBy = std::cos(2.0f*k*x(1));
            float sinCz = std::sin(3.0f*k*x(2));

            return sinAx*cosBy + (1.0f - cosAx)*(1.0f - sinBy) + sinCz*sinCz;
        };
        pde.sourceFunc = [&pde, dirichletFreq](const wos::Vector3& x) -> float {
            wos::Vector3 dAlpha = wos::Vector3::Zero();
            float alpha = pde.diffusionGradient(x, dAlpha);
            float sigma = pde.absorption(x);
            float a = wos::M_PI*dirichletFreq;
            float b = 2.0f*a;
            float c = 3.0f*a;
            float sinAx = std::sin(a*x(0));
            float cosAx = std::cos(a*x(0));
            float sinBy = std::sin(b*x(1));
            float cosBy = std::cos(b*x(1));
            float sinCz = std::sin(c*x(2));
            float cosCz = std::cos(c*x(2));

            float u = sinAx*cosBy + (1.0f - cosAx)*(1.0f - sinBy) + sinCz*sinCz;
            wos::Vector3 du = wos::Vector3::Zero();
            du(0) = (cosAx*cosBy + sinAx*(1.0f - sinBy))*a;
            du(1) = -(sinAx*sinBy + (1.0f - cosAx)*cosBy)*b;
            float d2udx2 = (cosAx*(1.0f - sinBy) - sinAx*cosBy)*a*a;
            float d2udy2 = ((1.0f - cosAx)*sinBy - sinAx*cosBy)*b*b;
            float d2u = d2udx2 + d2udy2;
            du(2) = 2.0f*sinCz*cosCz*c;
            float d2udz2 = 2.0f*(cosCz*cosCz - sinCz*sinCz)*c*c;
            d2u += d2udz2;

            return -alpha*d2u - dAlpha.dot(du) + sigma*u;
        };

        pde.bound = compute_bound(pde, {-1.0, -1.0, -1.0},{1.0, 1.0, 1.0});
    }

    // Generate points we want to compute the solution at.
    const int imageWidth = 256;
    const int imageHeight = 256;
    std::vector<wos::Vector3> pts = generate_slice_plane_points(
        wos::Vector3(-1, -1, 0), wos::Vector3(1, 1, 0), 0.01f,
        imageWidth, imageHeight
    );

    // Create the solver.
    wos::WalkOnSpheres<float, 3> solver(
        domain_bunny,                       // Which scene to use, completely defined by an SDF
        1000                                // Maximum number of steps
    );


    // Run the solver.
    std::vector<wos::SampleStatistics<float, 3>> results;
    solver.solve(
        pts,                                // The points we want to compute the solution at
        pde,                                // The pde we want to solve
        wos::Algorithm::DeltaTracking,      // Use the delta-tracking variant of our algorithms
        wos::WeightWindow(0.8f, 1.2f),      // Use russian roulette and splitting
        150,                                // Number of walks per point
        0.001f,                             // Boundary epsilon
        0.0f,                               // Initial solution value, generally 0
        results                             // A place to store the solution and additional statistics for each point
    );

    // Compute solution min and max for visualization.
    float min_solution =  FLT_MAX;
    float max_solution = -FLT_MAX;
    for(const auto& r : results) {
        min_solution = std::min(min_solution, r.getSolution());
        max_solution = std::max(max_solution, r.getSolution());
    }


    // Write out a PPM file (code courtesy of https://raytracing.github.io/books/RayTracingInOneWeekend.html#outputanimage/theppmimageformat).
    std::cout << "P3\n" << imageWidth << ' ' << imageHeight << "\n255\n";
    for (int j = imageHeight-1; j >= 0; --j) {
        for (int i = 0; i < imageWidth; ++i) {
            // Scale solution to 0-1 range
            float sol = std::min(1.0f, std::max(0.0f, 
                (results[j + imageWidth*i].getSolution() - min_solution) / (max_solution - min_solution)));
            
            // Map to color
            wos::Vector3 rgb = viridis(sol);

            int ir = static_cast<int>(255.999 * rgb.x());
            int ig = static_cast<int>(255.999 * rgb.y());
            int ib = static_cast<int>(255.999 * rgb.z());
            std::cout << ir << ' ' << ig << ' ' << ib << '\n';
        }
    }

}


std::vector<wos::Vector3> generate_slice_plane_points(wos::Vector3 min, wos::Vector3 max, float z, int width, int height) {
    std::vector<wos::Vector3> pts;
    pts.reserve(width*height);

    for(int y = 0; y < height; y++) {
        for(int x = 0; x < width; x++) {
            wos::Vector3 pos = min.array() + (max-min).array() * wos::Vector3(float(x) / width, float(y) / height, 0).array();
            pts.push_back(pos);
        }
    }

    return pts;
}

template<typename T, int DIM>
float compute_bound(const wos::PDE<T, DIM> pde, const wos::Vector<DIM>& bmin, const wos::Vector<DIM>& bmax) {
    float boundMin = FLT_MAX, boundMax = -FLT_MAX;
    for(int i = 0; i < 10000; i++) {
        wos::Vector<DIM> p = bmin.array() + wos::sampleUniformVector<DIM>().array() * (bmax - bmin).array();
        auto v = pde.transformedAbsorption(p);
        boundMin = std::min(boundMin, v);
        boundMax = std::max(boundMax, v);
    }

    return boundMax - boundMin;
}


//////////////////////////////////////////////////////////////////
//// Analytic SDF scene code https://www.shadertoy.com/view/wtVyWK
//////////////////////////////////////////////////////////////////

//CC0 1.0 Universal https://creativecommons.org/publicdomain/zero/1.0/
//To the extent possible under law, Blackle Mori has waived all copyright and related or neighboring rights to this work.

//like the last few shaders I've published, this is a neural network experiment.
//this time, the neural network reproduces an SDF given the x,y,z coordinates
//using siren networks with skip connections: https://vsitzmann.github.io/siren/

using vec3 = Eigen::Vector3f;
using vec4 = Eigen::Vector4f;

template<typename T>
auto dot(T a, T b) {
    return a.dot(b);
}

template<typename T>
float length(T v) {
    return v.norm();
}

template<>
float length(float v) {
    return v;
}

Eigen::Vector4f sin(Eigen::Vector4f v) {
    return v.array().sin();
}


Eigen::Matrix4f mat4(
    double f00, double f01, double f02, double f03, 
    double f04, double f05, double f06, double f07, 
    double f08, double f09, double f10, double f11, 
    double f12, double f13, double f14, double f15) {
    Eigen::Matrix4f res;
    res <<   (float)f00,  (float)f01,  (float)f02,  (float)f03, 
             (float)f04,  (float)f05,  (float)f06,  (float)f07, 
             (float)f08,  (float)f09,  (float)f10,  (float)f11, 
             (float)f12,  (float)f13,  (float)f14,  (float)f15;
    return res.transpose();
}

float domain_bunny(vec3 p) {
    vec3 op = p;
    p.x() = op.z();
    p.y() = op.y();
    p.z() = op.x();
    p *= 0.75;

    //sdf is undefined outside the unit sphere, uncomment to witness the abominations
    if (length(p) > 1.) {
        return -length(p) / 0.75f;
    }
    //neural networks can be really compact... when they want to be
    vec4 f00=sin(p.y()*vec4(-3.02f,1.95f,-3.42f,-.60f)+p.z()*vec4(3.08f,.85f,-2.25f,-.24f)-p.x()*vec4(-.29f,1.16f,-3.74f,2.89f)+vec4(-.71f,4.50f,-3.24f,-3.50f));
    vec4 f01=sin(p.y()*vec4(-.40f,-3.61f,3.23f,-.14f)+p.z()*vec4(-.36f,3.64f,-3.91f,2.66f)-p.x()*vec4(2.90f,-.54f,-2.75f,2.71f)+vec4(7.02f,-5.41f,-1.12f,-7.41f));
    vec4 f02=sin(p.y()*vec4(-1.77f,-1.28f,-4.29f,-3.20f)+p.z()*vec4(-3.49f,-2.81f,-.64f,2.79f)-p.x()*vec4(3.15f,2.14f,-3.85f,1.83f)+vec4(-2.07f,4.49f,5.33f,-2.17f));
    vec4 f03=sin(p.y()*vec4(-.49f,.68f,3.05f,.42f)+p.z()*vec4(-2.87f,.78f,3.78f,-3.41f)-p.x()*vec4(-2.65f,.33f,.07f,-.64f)+vec4(-3.24f,-5.90f,1.14f,-4.71f));
    vec4 f10=sin(mat4(-.34,.06,-.59,-.76,.10,-.19,-.12,.44,.64,-.02,-.26,.15,-.16,.21,.91,.15)*f00+
        mat4(.01,.54,-.77,.11,.06,-.14,.43,.51,-.18,.08,.39,.20,.33,-.49,-.10,.19)*f01+
        mat4(.27,.22,.43,.53,.18,-.17,.23,-.64,-.14,.02,-.10,.16,-.13,-.06,-.04,-.36)*f02+
        mat4(-.13,.29,-.29,.08,1.13,.02,-.83,.32,-.32,.04,-.31,-.16,.14,-.03,-.20,.39)*f03+
        vec4(.73f,-4.28f,-1.56f,-1.80f))/1.0f+f00;
    vec4 f11=sin(mat4(-1.11,.55,-.12,-1.00,.16,.15,-.30,.31,-.01,.01,.31,-.42,-.29,.38,-.04,.71)*f00+
        mat4(.96,-.02,.86,.52,-.14,.60,.44,.43,.02,-.15,-.49,-.05,-.06,-.25,-.03,-.22)*f01+
        mat4(.52,.44,-.05,-.11,-.56,-.10,-.61,-.40,-.04,.55,.32,-.07,-.02,.28,.26,-.49)*f02+
        mat4(.02,-.32,.06,-.17,-.59,.00,-.24,.60,-.06,.13,-.21,-.27,-.12,-.14,.58,-.55)*f03+
        vec4(-2.24f,-3.48f,-.80f,1.41f))/1.0f+f01;
    vec4 f12=sin(mat4(.44,-.06,-.79,-.46,.05,-.60,.30,.36,.35,.12,.02,.12,.40,-.26,.63,-.21)*f00+
        mat4(-.48,.43,-.73,-.40,.11,-.01,.71,.05,-.25,.25,-.28,-.20,.32,-.02,-.84,.16)*f01+
        mat4(.39,-.07,.90,.36,-.38,-.27,-1.86,-.39,.48,-.20,-.05,.10,-.00,-.21,.29,.63)*f02+
        mat4(.46,-.32,.06,.09,.72,-.47,.81,.78,.90,.02,-.21,.08,-.16,.22,.32,-.13)*f03+
        vec4(3.38f,1.20f,.84f,1.41f))/1.0f+f02;
    vec4 f13=sin(mat4(-.41,-.24,-.71,-.25,-.24,-.75,-.09,.02,-.27,-.42,.02,.03,-.01,.51,-.12,-1.24)*f00+
        mat4(.64,.31,-1.36,.61,-.34,.11,.14,.79,.22,-.16,-.29,-.70,.02,-.37,.49,.39)*f01+
        mat4(.79,.47,.54,-.47,-1.13,-.35,-1.03,-.22,-.67,-.26,.10,.21,-.07,-.73,-.11,.72)*f02+
        mat4(.43,-.23,.13,.09,1.38,-.63,1.57,-.20,.39,-.14,.42,.13,-.57,-.08,-.21,.21)*f03+
        vec4(-.34f,-3.28f,.43f,-.52f))/1.0f+f03;
    f00=sin(mat4(-.72,.23,-.89,.52,.38,.19,-.16,-.88,.26,-.37,.09,.63,.29,-.72,.30,-.95)*f10+
        mat4(-.22,-.51,-.42,-.73,-.32,.00,-1.03,1.17,-.20,-.03,-.13,-.16,-.41,.09,.36,-.84)*f11+
        mat4(-.21,.01,.33,.47,.05,.20,-.44,-1.04,.13,.12,-.13,.31,.01,-.34,.41,-.34)*f12+
        mat4(-.13,-.06,-.39,-.22,.48,.25,.24,-.97,-.34,.14,.42,-.00,-.44,.05,.09,-.95)*f13+
        vec4(.48f,.87f,-.87f,-2.06f))/1.4f+f10;
    f01=sin(mat4(-.27,.29,-.21,.15,.34,-.23,.85,-.09,-1.15,-.24,-.05,-.25,-.12,-.73,-.17,-.37)*f10+
        mat4(-1.11,.35,-.93,-.06,-.79,-.03,-.46,-.37,.60,-.37,-.14,.45,-.03,-.21,.02,.59)*f11+
        mat4(-.92,-.17,-.58,-.18,.58,.60,.83,-1.04,-.80,-.16,.23,-.11,.08,.16,.76,.61)*f12+
        mat4(.29,.45,.30,.39,-.91,.66,-.35,-.35,.21,.16,-.54,-.63,1.10,-.38,.20,.15)*f13+
        vec4(-1.72f,-.14f,1.92f,2.08f))/1.4f+f11;
    f02=sin(mat4(1.00,.66,1.30,-.51,.88,.25,-.67,.03,-.68,-.08,-.12,-.14,.46,1.15,.38,-.10)*f10+
        mat4(.51,-.57,.41,-.09,.68,-.50,-.04,-1.01,.20,.44,-.60,.46,-.09,-.37,-1.30,.04)*f11+
        mat4(.14,.29,-.45,-.06,-.65,.33,-.37,-.95,.71,-.07,1.00,-.60,-1.68,-.20,-.00,-.70)*f12+
        mat4(-.31,.69,.56,.13,.95,.36,.56,.59,-.63,.52,-.30,.17,1.23,.72,.95,.75)*f13+
        vec4(-.90f,-3.26f,-.44f,-3.11f))/1.4f+f12;
    f03=sin(mat4(.51,-.98,-.28,.16,-.22,-.17,-1.03,.22,.70,-.15,.12,.43,.78,.67,-.85,-.25)*f10+
        mat4(.81,.60,-.89,.61,-1.03,-.33,.60,-.11,-.06,.01,-.02,-.44,.73,.69,1.02,.62)*f11+
        mat4(-.10,.52,.80,-.65,.40,-.75,.47,1.56,.03,.05,.08,.31,-.03,.22,-1.63,.07)*f12+
        mat4(-.18,-.07,-1.22,.48,-.01,.56,.07,.15,.24,.25,-.09,-.54,.23,-.08,.20,.36)*f13+
        vec4(-1.11f,-4.28f,1.02f,-.23f))/1.4f+f13;
    return -(dot(f00,vec4(.09f,.12f,-.07f,-.03f))+dot(f01,vec4(-.04f,.07f,-.08f,.05f))+
        dot(f02,vec4(-.01f,.06f,-.02f,.07f))+dot(f03,vec4(-.05f,.07f,.03f,.04f))-0.16f) / 0.75f;
}


//////////////////////////////////////////////////////////////
// Colormap https://www.shadertoy.com/view/WlfXRN
//////////////////////////////////////////////////////////////
vec3 viridis(float t) {
    const vec3 c0 = vec3(0.2777273272234177f, 0.005407344544966578f, 0.3340998053353061f);
    const vec3 c1 = vec3(0.1050930431085774f, 1.404613529898575f, 1.384590162594685f);
    const vec3 c2 = vec3(-0.3308618287255563f, 0.214847559468213f, 0.09509516302823659f);
    const vec3 c3 = vec3(-4.634230498983486f, -5.799100973351585f, -19.33244095627987f);
    const vec3 c4 = vec3(6.228269936347081f, 14.17993336680509f, 56.69055260068105f);
    const vec3 c5 = vec3(4.776384997670288f, -13.74514537774601f, -65.35303263337234f);
    const vec3 c6 = vec3(-5.435455855934631f, 4.645852612178535f, 26.3124352495832f);
    return c0+t*(c1+t*(c2+t*(c3+t*(c4+t*(c5+t*c6)))));
}
//////////////////////////////////////////////////////////////