#pragma once

#include <algorithm>
#include <cmath>
#include <iostream>
#include <limits>
#include <memory>
#include <string>
#include <chrono>
#include <random>
#include <Eigen/Core>
#include <Eigen/Geometry>
#include "bessel.hpp"

namespace wos {

constexpr float M_PI = 3.141592653f;

template<int DIM>
using Vector = Eigen::Matrix<float, DIM, 1>;
using Vector2 = Vector<2>;
using Vector3 = Vector<3>;

inline float sampleUniform(float a=0.0f, float b=1.0f)
{
	thread_local std::mt19937 generator(std::random_device{}());
	std::uniform_real_distribution<float> distribution(a, b);

	return distribution(generator);
}

inline float sampleNormal(float mean=0.0f, float stddev=1.0f)
{
	thread_local std::mt19937 generator(std::random_device{}());
	std::normal_distribution<float> distribution(mean, stddev);

	return distribution(generator);
}

template <int DIM>
inline Vector<DIM> sampleUniformVector(float a=0.0f, float b=1.0f)
{
	Vector<DIM> v;
	for (int i = 0; i < DIM; i++) {
		v(i) = sampleUniform(a, b);
	}

	return v;
}

template <int DIM>
inline Vector<DIM> sampleNormalVector(float mean=0.0f, float stddev=1.0f)
{
	Vector<DIM> v;
	for (int i = 0; i < DIM; i++) {
		v(i) = sampleNormal(mean, stddev);
	}

	return v;
}

template <int DIM>
inline Vector<DIM> sampleUnitSphereUniform()
{
	Vector<DIM> v = sampleNormalVector<DIM>();

	return v/v.norm();
}

template <>
inline Vector2 sampleUnitSphereUniform<2>()
{
	float u = sampleUniform();
	float phi = 2.0f*M_PI*u;

	return Vector2(std::cos(phi), std::sin(phi));
}

template <>
inline Vector3 sampleUnitSphereUniform<3>()
{
	float u1 = sampleUniform();
	float u2 = sampleUniform();
	float z = 1.0f - 2.0f*u1;
	float r = std::sqrt(std::max(0.0f, 1.0f - z*z));
	float phi = 2.0f*M_PI*u2;

	return Vector3(r*std::cos(phi), r*std::sin(phi), z);
}

template <int DIM>
inline Vector<DIM> sampleUnitBallUniform()
{
	float u = sampleUniform();
	float r = std::pow(u, 1.0f/DIM);

	return r*sampleUnitSphereUniform<DIM>();
}

template <>
inline Vector2 sampleUnitBallUniform<2>()
{
	float u = sampleUniform();
	float r = std::sqrt(u);

	return r*sampleUnitSphereUniform<2>();
}

template <>
inline Vector3 sampleUnitBallUniform<3>()
{
	float u = sampleUniform();
	float r = std::cbrt(u);

	return r*sampleUnitSphereUniform<3>();
}

template <int DIM>
class GreensFnBall {
public:
	// constructor
	GreensFnBall() {
		updateBall(Vector<DIM>::Zero(), 0.0f);
	}

	// destructor
	virtual ~GreensFnBall() {}

	// updates the ball center and radius
	virtual void updateBall(const Vector<DIM>& c_, float R_) {
		c = c_;
		yVol = Vector<DIM>::Zero();
		ySurf = Vector<DIM>::Zero();
		R = R_;
		r = 0.0f;
	}

	// samples a point inside the ball
	virtual Vector<DIM> sampleVolume(float& pdf) {
		return Vector<DIM>::Zero();
	}

	// evaluates the Green's function
	virtual float evaluate() const {
		return 0.0f;
	}

	// evaluates the gradient of the Green's function
	virtual Vector<DIM> gradient() const {
		return Vector<DIM>::Zero();
	}

	// evaluates the norm of the Green's function
	virtual float norm() const {
		return 0.0f;
	}

	// samples a point on the surface of the ball
	virtual Vector<DIM> sampleSurface(float& pdf) {
		return Vector<DIM>::Zero();
	}

	// evaluates the Poisson Kernel (normal derivative of the Green's function)
	virtual float poissonKernel() const {
		return 0.0f;
	}

	// evaluates the gradient of the Poisson Kernel
	virtual Vector<DIM> poissonKernelGradient() const {
		return Vector<DIM>::Zero();
	}

	// members
	Vector<DIM> c, yVol, ySurf; // ball center, sampled point inside and on the surface of the ball
	float R, r; // ball radius and distance of the sampled point inside the ball to the ball center

protected:
	// samples a point inside the ball
	virtual Vector<DIM> rejectionSampleGreensFn(float bound, float& pdf) {
		// rejection sample
		int iter = 0;

		do {
			float u = sampleUniform();
			r = sampleUniform()*R;
			pdf = evaluate()/norm();
			float pdfRadius = (DIM == 2 ? 2.0f*M_PI*r : 4.0f*M_PI*r*r)*pdf;
			iter++;

			if (u < pdfRadius/bound) {
				break;
			}

		} while (iter < 1000);

		r = std::max(5e-7f, r);
		yVol = c + r*sampleUnitSphereUniform<DIM>();

		return yVol;
	}
};

template <int DIM>
class HarmonicGreensFnBall: public GreensFnBall<DIM> {
public:
	// constructor
	HarmonicGreensFnBall(): GreensFnBall<DIM>() {
		std::cerr << "HarmonicGreensFnBall() not implemented for DIM: " << DIM << std::endl;
		exit(EXIT_FAILURE);
	}
};

template <>
class HarmonicGreensFnBall<2>: public GreensFnBall<2> {
public:
	// constructor
	HarmonicGreensFnBall(): GreensFnBall<2>() {}

	// samples a point inside the ball
	Vector2 sampleVolume(float& pdf) {
		// TODO: can probably do better
		// rejection sample radius r from pdf 4.0 * r * ln(R / r) / R^2
		float bound = 1.5f/R;

		return GreensFnBall<2>::rejectionSampleGreensFn(bound, pdf);
	}

	// evaluates the Green's function
	float evaluate() const {
		return std::log(R/r)/(2.0f*M_PI);
	}

	// evaluates the gradient of the Green's function
	Vector2 gradient() const {
		Vector2 d = yVol - c;

		return d*(1.0f/(r*r) - 1.0f/(R*R))/(2.0f*M_PI);
	}

	// evaluates the norm of the Green's function
	float norm() const {
		return R*R/4.0f;
	}

	// samples a point on the surface of the ball
	Vector2 sampleSurface(float& pdf) {
		ySurf = c + R*sampleUnitSphereUniform<2>();
		pdf = 1.0f/(2.0f*M_PI);

		return ySurf;
	}

	// evaluates the Poisson Kernel (normal derivative of the Green's function)
	float poissonKernel() const {
		return 1.0f/(2.0f*M_PI);
	}

	// evaluates the gradient of the Poisson Kernel
	Vector2 poissonKernelGradient() const {
		Vector2 d = ySurf - c;

		return 2.0f*d/(2.0f*M_PI*R*R);
	}
};

template <>
class HarmonicGreensFnBall<3>: public GreensFnBall<3> {
public:
	// constructor
	HarmonicGreensFnBall(): GreensFnBall<3>() {}

	// samples a point inside the ball
	Vector3 sampleVolume(float& pdf) {
		// TODO: can probably do better
		// rejection sample radius r from pdf 6.0f * r * (R - r) / R^3
		float bound = 1.5f/R;

		return GreensFnBall<3>::rejectionSampleGreensFn(bound, pdf);
	}

	// evaluates the Green's function
	float evaluate() const {
		return (1.0f/r - 1.0f/R)/(4.0f*M_PI);
	}

	// evaluates the gradient of the Green's function
	Vector3 gradient() const {
		Vector3 d = yVol - c;

		return d*(1.0f/(r*r*r) - 1.0f/(R*R*R))/(4.0f*M_PI);
	}

	// evaluates the norm of the Green's function
	float norm() const {
		return R*R/6.0f;
	}

	// samples a point on the surface of the ball
	Vector3 sampleSurface(float& pdf) {
		ySurf = c + R*sampleUnitSphereUniform<3>();
		pdf = 1.0f/(4.0f*M_PI);

		return ySurf;
	}

	// evaluates the Poisson Kernel (normal derivative of the Green's function)
	float poissonKernel() const {
		return 1.0f/(4.0f*M_PI);
	}

	// evaluates the gradient of the Poisson Kernel
	Vector3 poissonKernelGradient() const {
		Vector3 d = ySurf - c;

		return 3.0f*d/(4.0f*M_PI*R*R);
	}
};

template <int DIM>
class YukawaGreensFnBall: public GreensFnBall<DIM> {
public:
	// constructor
	YukawaGreensFnBall(float lambda_): GreensFnBall<DIM>() {
		std::cerr << "YukawaGreensFnBall() not implemented for DIM: " << DIM << std::endl;
		exit(EXIT_FAILURE);
	}
};

template <>
class YukawaGreensFnBall<2>: public GreensFnBall<2> {
public:
	// constructor
	YukawaGreensFnBall(float lambda_): GreensFnBall<2>(), lambda(lambda_), sqrtLambda(std::sqrt(lambda_)) {}

	// updates the ball center and radius
	void updateBall(const Vector2& c_, float R_) {
		GreensFnBall<2>::updateBall(c_, R_);
		muR = R*sqrtLambda;
		K0muR = (float)bessel::bessk0(muR);
		I0muR = (float)bessel::bessi0(muR);
		K1muR = (float)bessel::bessk1(muR);
		I1muR = (float)bessel::bessi1(muR);
	}

	// samples a point inside the ball
	Vector2 sampleVolume(float& pdf) {
		// TODO: can probably do better
		// rejection sample radius r from pdf r * λ * (K_0(r√λ) * I_0(R√λ) - I_0(r√λ) * K_0(R√λ)) / (I_0(R√λ) - 1)
		float bound = R <= lambda ? std::max(std::max(2.2f/R, 2.2f/lambda), std::max(0.6f*std::sqrt(R), 0.6f*sqrtLambda)) :
																std::max(std::min(2.2f/R, 2.2f/lambda), std::min(0.6f*std::sqrt(R), 0.6f*sqrtLambda));

		return GreensFnBall<2>::rejectionSampleGreensFn(bound, pdf);
	}

	// evaluates the Green's function
	float evaluate() const {
		double mur = r*sqrtLambda;
		double K0mur = bessel::bessk0(mur);
		double I0mur = bessel::bessi0(mur);

		return float(K0mur - (I0mur/I0muR)*K0muR)/(2.0f*M_PI);
	}

	// evaluates the gradient of the Green's function
	Vector2 gradient() const {
		Vector2 d = yVol - c;
		float mur = r*sqrtLambda;
		float K1mur = (float)bessel::bessk1(mur);
		float I1mur = (float)bessel::bessi1(mur);
		float Qr = sqrtLambda*(K1mur - I1mur*K1muR/I1muR);

		return d*Qr/(2.0f*M_PI*r);
	}

	// evaluates the norm of the Green's function
	float norm() const {
		return (1.0f - 2.0f*M_PI*poissonKernel())/lambda;
	}

	// samples a point on the surface of the ball
	Vector2 sampleSurface(float& pdf) {
		ySurf = c + R*sampleUnitSphereUniform<2>();
		pdf = 1.0f/(2.0f*M_PI);

		return ySurf;
	}

	// evaluates the Poisson Kernel (normal derivative of the Green's function)
	float poissonKernel() const {
		return 1.0f/(2.0f*M_PI*I0muR);
	}

	// evaluates the gradient of the Poisson Kernel
	Vector2 poissonKernelGradient() const {
		Vector2 d = ySurf - c;
		float QR = sqrtLambda/(R*I1muR);

		return d*QR/(2.0f*M_PI);
	}

private:
	// members
	float lambda, sqrtLambda;
	float muR, K0muR, I0muR, K1muR, I1muR;
};

template <>
class YukawaGreensFnBall<3>: public GreensFnBall<3> {
public:
	// constructor
	YukawaGreensFnBall(float lambda_): GreensFnBall<3>(), lambda(lambda_), sqrtLambda(std::sqrt(lambda_)) {}

	// updates the ball center and radius
	void updateBall(const Vector3& c_, float R_) {
		GreensFnBall<3>::updateBall(c_, R_);
		muR = R*sqrtLambda;
		expmuR = std::exp(-muR);
		float exp2muR = expmuR*expmuR;
		float coshmuR = (1.0f + exp2muR)/(2.0f*expmuR);
		sinhmuR = (1.0f - exp2muR)/(2.0f*expmuR);
		K32muR = expmuR*(1.0f + 1.0f/muR);
		I32muR = coshmuR - sinhmuR/muR;
	}

	// samples a point inside the ball
	Vector3 sampleVolume(float& pdf) {
		// TODO: can probably do better
		// rejection sample radius r from pdf r * λ * sinh((R - r)√λ) / (sinh(R√λ) - R√λ)
		float bound = R <= lambda ? std::max(std::max(2.0f/R, 2.0f/lambda), std::max(0.5f*std::sqrt(R), 0.5f*sqrtLambda)) :
																std::max(std::min(2.0f/R, 2.0f/lambda), std::min(0.5f*std::sqrt(R), 0.5f*sqrtLambda));

		return GreensFnBall<3>::rejectionSampleGreensFn(bound, pdf);
	}

	// evaluates the Green's function
	float evaluate() const {
		float mur = r*sqrtLambda;
		float expmur = std::exp(-mur);
		float sinhmur = (1.0f - expmur*expmur)/(2.0f*expmur);

		return (expmur - expmuR*sinhmur/sinhmuR)/(4.0f*M_PI*r);
	}

	// evaluates the gradient of the Green's function
	Vector3 gradient() const {
		Vector3 d = yVol - c;
		float mur = r*sqrtLambda;
		float expmur = std::exp(-mur);
		float exp2mur = expmur*expmur;
		float coshmur = (1.0f + exp2mur)/(2.0f*expmur);
		float sinhmur = (1.0f - exp2mur)/(2.0f*expmur);
		float K32mur = expmur*(1.0f + 1.0f/mur);
		float I32mur = coshmur - sinhmur/mur;
		float Qr = sqrtLambda*(K32mur - I32mur*K32muR/I32muR);

		return d*Qr/(4.0f*M_PI*r*r);
	}

	// evaluates the norm of the Green's function
	float norm() const {
		return (1.0f - 4.0f*M_PI*poissonKernel())/lambda;
	}

	// samples a point on the surface of the ball
	Vector3 sampleSurface(float& pdf) {
		ySurf = c + R*sampleUnitSphereUniform<3>();
		pdf = 1.0f/(4.0f*M_PI);

		return ySurf;
	}

	// evaluates the Poisson Kernel (normal derivative of the Green's function)
	float poissonKernel() const {
		return muR/(4.0f*M_PI*sinhmuR);
	}

	// evaluates the gradient of the Poisson Kernel
	Vector3 poissonKernelGradient() const {
		Vector3 d = ySurf - c;
		float QR = lambda/I32muR;

		return d*QR/(4.0f*M_PI);
	}

private:
	// members
	float lambda, sqrtLambda; // potential
	float muR, expmuR, sinhmuR, K32muR, I32muR;
};

} // mc
